--[[

    File: santa.lua 

    Copyright (C) 2000-2008 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

--]]


require 'common.pingpong'
require 'common.hsv'
require 'common.shader'

--[[
accomodating for time zones ...
1) time zones, color each separately and distinctly, and provide a mapping from each color to its appropriate GST offset
2) next count up time zone map histogram and calc weights of each time zone appropriately
2.5) divy up the curve vertices according to proportionate area of each time zone ...
	bias towards higher coastline ... cuz islands still need to be covered too, though they wont' come up as often
3) when picking a point, find what time zone it is in, then pick the best portion of the curve that pertains to the time zone
--]]

--[[
logGaussianMu is our exp coefficient ...
we want to redo our exp to be texture size independent
--]]

if glush then
	glush.gui.slider('convergenceRate', 0, 1, 1)
	glush.gui.slider('logGaussianMu', -10, -5, 10)
	glush.gui.checkbox('doUpdate', false)
	glush.gui.checkbox('doSingleUpdate', false)
	glush.gui.checkbox('doReset', false)
--	glush.gui.checkbox('doDetailUp', false)
--	glush.gui.checkbox('doDetailDown', false)
else
	convergenceRate = .75
	logGaussianMu = 0
end

glush.cvarExec('zNear .001')
glush.cvarExec('zFar 10')

hsvTex = HsvTex.new(256)

function latticeInitTexel()
	return math.random(), math.random(), 0, 1
end

function createFloatPingPong(w,h)
	return PingPong.new{
		width=w;
		height=h;
		minFilter=gl.GL_NEAREST;
		magFilter=gl.GL_NEAREST;
		format=gl.GL_RGBA;
		internalFormat=gl.GL_RGBA_FLOAT32;
		data=latticeInitTexel;
	}
end

sphereProjectCode = [[
vec3 sphereProject(vec2 v) {
	const float pi = ]]..math.pi..[[;
	float th = 2. * pi * v.x;
	float ph = pi * (v.y - .5);
	float cosph = cos(ph);
	return vec3(cosph * cos(th), cosph * sin(th), -sin(ph));
}
]]

sphereUnprojectCode = [[
vec2 sphereUnproject(vec3 v) {
	const float pi = ]]..math.pi..[[;
	float th = atan(v.y, v.x);
	float xy = length(v.xy);
	float ph = atan(-v.z, xy);
	vec2 res;
	res.x = th / (2. * pi);
	res.x = fract(res.x);
	res.y = ph / pi + .5;
	res.y = clamp(res.y, 0., 1.);
	return res;
}
]]

errorFunctionCode = sphereProjectCode..[[
float errorFunction(vec4 x, vec4 y) {
	vec3 u = sphereProject(x.xy);
	vec3 v = sphereProject(y.xy);
	return 1. - dot(u,v);
}
]]

latticeTexelDistSqCode = [[
float latticeTexelDistSq(vec2 a, vec2 b) {
	//2D euclidian
	//vec2 del = a - b;
	//return dot(del,del);
	//2D unwrapped to 3D ...
	//this isn't unwrapping correctly...
	float del = (a.x - b.x) / texsize.y + a.y - b.y;
	del = min(del, 1. - del);
	return del * del;
	//del = sin(]]..math.pi..[[ * del);	//toric
	//return del * del;
}
]]

-- takes two points in the lattice's element's vector space
-- lerps them
-- returns the result
latticeVectorLerpCode = sphereProjectCode..sphereUnprojectCode..[[
vec4 latticeVectorLerp(vec4 a, vec4 b, float infl) {
	vec3 u = sphereProject(a.xy);
	vec3 v = sphereProject(b.xy);
#if 0
	// now for a slerp ... which means cross product to get the axis, rotate according to the interpolant	
	vec3 axis = cross(u,v);
	axis = normalize(axis);
	float angle = acos(clamp(dot(u,v),-1.,1.));
	angle *= infl;
	//rotate u along axis by angle radians
	vec3 dir = cross(axis, u);	//should already be unit since axis and u are right angles and both are units
	vec3 w = u * cos(angle) + dir * sin(angle);
	//and then reconstruct ...
#else
	//or just lerp and wing it ...
	vec3 w = mix(u,v,infl);
#endif
	vec4 res;
	res.xy = sphereUnproject(w);
	return res;
}
]]

calcErrorShader = KernelShader.new{
	texs={'tex'};
	uniforms={inputVector='vec4';};
	code=errorFunctionCode..[[ 
void main() {
	vec4 texel = texture2D(tex, pos);
	float e = errorFunction(inputVector, texel);
	gl_FragColor = vec4(pos.x, pos.y, e, 0);
}
]];
}

findBestShader = KernelShader.new{
	texs={'tex'};
	uniforms={viewportsize='vec2'; texsize='vec2';};
	code= [[
void main() {
	//pos is in viewport coords
	vec2 tc = pos * viewportsize / texsize;
	//tc is in tex coords
	vec2 dtc = 1. / texsize;
	vec2 dp = 2. * tc - .5 * dtc;

	vec4 s1 = texture2D(tex, dp);
	vec4 s2 = texture2D(tex, dp + vec2(dtc.x, 0.));
	vec4 s3 = texture2D(tex, dp + vec2(0., dtc.y));
	vec4 s4 = texture2D(tex, dp + vec2(dtc.x, dtc.y));
	
	vec4 best = s1;
	if (s2.z < best.z) best = s2;
	if (s3.z < best.z) best = s3;
	if (s4.z < best.z) best = s4;
	gl_FragColor = best;
}
]];
}

pullMapShader = KernelShader.new{
	texs={'tex'};
	uniforms={
		inputVector='vec4';
		bestPos='vec2';
		texsize='vec2';
		mu='float';
		inflCoeff='float';
	};
	code= latticeVectorLerpCode..
latticeTexelDistSqCode..[[ 
void main() {
	vec4 texel = texture2D(tex, pos);
	float delSq = latticeTexelDistSq(pos, bestPos);
	float infl = exp(-delSq/(mu*mu)) * inflCoeff;
	
	//adjustment needs to be in the space of the vectors ...
	gl_FragColor = latticeVectorLerp(texel, inputVector, infl);

	//color goods for 1D map
	gl_FragColor.w = pos.x / texsize.y + pos.y;
}
]];
}

displayShader = ShaderProgram.new{
	vertexCode = 
sphereProjectCode..[[
varying float color;
void main() {
	color = gl_Vertex.w;
	vec3 v3 = sphereProject(gl_Vertex.xy);
	vec4 v = vec4(v3, 1.);
	gl_Position = gl_ModelViewProjectionMatrix * v;
}
]];
	fragmentCode = [[
varying float color;
uniform sampler1D hsvTex;
void main() {
	gl_FragColor = texture1D(hsvTex, color);
}
]];
	uniforms = {
		hsvTex = 0;
	};
}

Lattice = {}
Lattice.__index = Lattice

function Lattice:new(width, height)
	local self = {}
	setmetatable(self, Lattice)
	self.width = width
	self.height = height
	self.components = 4	-- asserted by createFloatTex's contents.
	
	-- the base texture
	self.texs = createFloatPingPong(self.width, self.height)

	-- what holds our error calcs
	self.error = createFloatPingPong(self.width, self.height)

	return self
end

function Lattice:calcError(input)
--	v.error = (v - x):length()
	local bestPos
	
	-- this stores [x,y,error] across the texture
	self.error:draw{
		shader = calcErrorShader;
		texs = { self.texs:cur() };
		uniforms = {
			inputVector = input;
		};
		viewport = { 0, 0, self.width, self.height };
		callback = function()
			drawScreenQuad()
			-- get a default bestPos in case we start off in a 1x1 state
			bestPos = gl.glReadPixels(0,0,1,1)
		end;
	}
	
--	do return {0,0} end
	
	-- now we ping-pong and reduce (between error and swapTexID since they're the same size)
	local dw = self.width
	local dh = self.height
	while dw > 1 or dh > 1 do
		dw = dw / 2
		if dw < 1 then dw = 1 end
		dh = dh / 2
		if dh < 1 then dh = 1 end
		self.error:swap()	-- I don't see a need for a swap here...
		self.error:draw{
			shader = findBestShader;
			texs = { self.error:prev() };
			viewport = { 0, 0, dw, dh };
			uniforms = {
				viewportsize = {dw,dh};
				texsize = {self.width, self.height};
			};
			callback = function()
				drawScreenQuad()
				if dw == 1 and dh == 1 then
					bestPos = gl.glReadPixels(0,0,1,1)
				end
--[[
				--debug: spit it all out
				local t = gl.glReadPixels(0,0,dw,dh)
				setmetatable(t, Lattice)
				print('best so far for size '..dw..', '..dh..': '..#t)
				local bestIndex, bestError
				for i=1,#t/4 do
					if not bestIndex or bestError > t[4*i-1] then
						bestError = t[4*i-1]
						bestIndex = {t[4*i-3], t[4*i-2]}
					end
				end
				print('best error '..bestError)
				print('best index '..bestIndex[1]..', '..bestIndex[2])
--]]
--[[
				for i=1,#t do
					if i % 4 == 1 then io.write('{')
					else io.write(', ') end
					io.write(t[i])
					if i % 4 == 0 then io.write('} ') end
				end
				io.write('\n')
--]]
			end;
		}
	end
	
	return bestPos
end

-- bestError is in texel coords, so you'll have to normalize them before using them
-- gen a gaussian map from the best error ...
--		input: i,j of best point - as a uniform
--		output: fill G component of error with exp(-||(i,j)-(x,y)||^2/mu^2).  preserve R
-- pull the map towards it
--		input: old tex, error's G component holding the gauss map, convergenceCoefficient uniform
--		output: new tex = old tex + (x - old tex) * error's G component * convergenceCoefficient
function Lattice:pullMap(input, bestPos)
	self.texs:swap()
	self.texs:draw{
		shader = pullMapShader;
		viewport = { 0, 0, self.width, self.height };
		texs = { self.texs:prev() };
		uniforms = {
			inputVector = input;
			bestPos = {bestPos[1], bestPos[2]};
			mu = math.exp(logGaussianMu);
			inflCoeff = convergenceRate;
			texsize = {self.width, self.height};
		};
	}
end

function Lattice:getArray()
	gl.glBindTexture(gl.GL_TEXTURE_2D, self.texs:cur().id)
	local t = gl.glGetTexImage(gl.GL_TEXTURE_2D, 0)
	gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
	return t
end

function Lattice:getErrorArray()
	gl.glBindTexture(gl.GL_TEXTURE_2D, self.error:cur().id)
	local t = gl.glGetTexImage(gl.GL_TEXTURE_2D, 0)
	gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
	return t
end

-- seems like I should hijack the concat metatable or something
function Lattice:tostring()
	local t = self:getArray()

	local s = '{'
	local sep = ''
	for i,v in ipairs(t) do
		s = s .. sep .. v
		sep = ','
	end
	s = s .. '}'
	return s
end
Lattice.__tostring = Lattice.tostring

function Lattice:reset()
	gl.glBindTexture(gl.GL_TEXTURE_2D, self.texs:cur().id);
	gl.glTexSubImage2D(gl.GL_TEXTURE_2D, 0, 0, 0, self.width, self.height, gl.GL_RGBA, gl.GL_FLOAT, latticeInitTexel)
	gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
end

do
	src = {}
	src.tex = Tex2D.load('santa/land.png')
	--src.tex = glush.createTex2D('santa/land.png')
	src.visibleTex = Tex2D.load('santa/world.png')
	--src.width = glGetTexLevelParameter(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH)
	--src.height = glGetTexLevelParameter(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT)
	assert(src.tex.width > 0 and src.tex.height > 0, "ut oh we can't make a FBO out of this")
	
	src.tex:bind()
	src.buffer = gl.glGetTexImage(gl.GL_TEXTURE_2D, 0)
	src.tex:unbind()
	src.components = #src.buffer / (src.tex.width * src.tex.height)
end

do
	--[[
	TODO: progressive refinement
	start at 4x4
	run it N iterations
	(where N is ...
		... fixed?
		or proporitional to the current lattice size?)
	then subdivide (interpolate the new texels)
	a b
	c d
		=>
	a 0 0 0
	b 0 0 0
	c 0 0 0
	d 0 0 0

	and keep going
	
	always keeping the mu just right to account for three texels or so on either side of the pick

	also train level M by a texture downfiltered to MxM (or some appropriate low resolution)

	OR

	we can skip the subdivide/resampling
	and just gradually adjust M as we want to successively refine the path
	and it should ahve the same effect
	just with a bit of extra computations - from those intermediate points we drag around at first
	
	--]]
	l = Lattice:new(128,128)	-- the product being the number of classes for our image
	
	-- we also want a pbo/vbo with our lattice float ...
	-- but that would mean ensuring the lattice always ended our texture in the same spot
	-- or just re-uploading it every time.  no worries.
	
	-- the classes are all messed up. reorganize me plz.
	
	assert(l.components == 4)
	-- create the vbo
	l.vbo = gl.glGenBuffer()
	gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER, l.vbo)
	gl.glBufferData(gl.GL_PIXEL_PACK_BUFFER, l.width * l.height * l.components * 4, nil, gl.GL_DYNAMIC_DRAW)	-- # vtxs * components * sizeof(float)
	gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER, 0)

	gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER, l.vbo)
	gl.glBindTexture(gl.GL_TEXTURE_2D, l.texs:cur().id)
	gl.glGetTexImage(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, gl.GL_FLOAT, 0)
	gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
	gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER, 0)
end

--[[
balance the randomness of 'v' ...
sphere parameterization:
ph <- [-pi/2, pi/2]
th <- [0, 2*pi]
rho * (cos(ph) * cos(th), cos(ph) * sin(th), -sin(ph))
jacobian
[ cos(ph) * cos(th),  -sin(ph) * cos(th), -cos(ph) * sin(th)]
[ cos(ph) * sin(th), -sin(ph) * sin(th), cos(ph) * cos(th)]
[ -sin(ph), -cos(ph), 0]
determinant
-sin(ph) * ( -sin(ph) * cos(th) * cos(ph) * cos(th) - sin(ph) * sin(th) * cos(ph) * sin(th)  )  - cos(ph) * (cos(ph) * cos(th) * cos(ph) * cos(th) + cos(ph) * sin(th) * cos(ph) * sin(th) )
-sin(ph) * ( -sin(ph) * cos(ph) * cos(th)^2 - sin(ph) * cos(ph) * sin(th)^2 ) + cos(ph) * (cos(ph)^2 * cos(th)^2 + cos(ph)^2 * sin(th)^2)
-sin(ph) * ( -sin(ph) * cos(ph) ) + cos(ph) * cos(ph)^2
sin(ph)^2 * cos(ph) + cos(ph) * cos(ph)^2
cos(ph)

our distribution is cos(ph) from ph [-pi/2,pi/2]
remapping that to our u,v coordinates of [0,1]^2 ...
our distribution is f(v) = cos( (v-.5)*pi ) for v <- [0,1]
normalizing it ...
n(v) = f(v) / l(n)
l(n) = int(0,1, f(v) dv)
l(n) = int(0,1, cos((v-.5)*pi) dv)
	ph = (v-.5)*pi
	dph = pi dv
	dv = dph / pi
int(-pi/2, pi/2, cos(ph)/pi dph)
sin(pi/2)/pi - sin(-pi/2)/pi
1/pi - (-1/pi)
2/pi
so our normalized function is n(v) = cos( (v-.5)*pi )*pi/2
now we want a function that maps a fair [0,1] pick to our normalized function 
y = fair [0,1] pick, y' is our biased pick (that we want a map to)
int(0,y', n(v) dv) = y
N(y') - N(0) = y (for d/dv N(v) = n(v))
N(y') = y + N(0)
y' = N^-1(y + N(0))

now to solve N(v) ...
N(v) = int(0,v, cos( (v-.5)*pi)*pi/2 dv)
	ph = (v-.5)*pi
	dph = pi dv
	dv = dph / pi
N(v) = int(-pi/2, (v-.5)*pi, cos(ph)/pi dph)
N(v) = (sin( (v-.5)*pi ) + 1)/pi

OOPS I WAS SUPPOSED TO NORMALIZE THE ANTIDERIVATIVE!
N(v) = (sin( (v-.5)*pi ) + 1)/2

N(0) = (sin( -pi/2) + 1)/2
	= (-1 + 1)/2 = 0/2 = 0

N(1) = (sin( pi/2 ) + 1)/2
	= (1+1)/2 = 1

N = (sin( (v-.5)*pi ) + 1)/2
N*2 = sin( (v-.5)*pi ) + 1
N*2 - 1 = sin( (v-.5)*pi )
sin^-1(N*2 - 1) = (v - .5)*pi
sin^-1(N*2 - 1)/pi + .5 = v

N^-1(v) = sin^-1(2*v - 1)/pi + .5

--]]
function inverseSphereDistribute(v)
	return math.asin(2 * v - 1)/math.pi + .5
end

function update()
	local u, v
	while true do
		u = math.random() * src.tex.width
		v = inverseSphereDistribute(math.random()) * src.tex.height
		local index = math.floor(u) + math.floor(v) * src.tex.width
		if src.buffer[src.components * index + 1] > .5 then break end
	end
	local input = { (u+.5)/src.tex.width, (v+.5)/src.tex.height, 0, 1}
	-- run the error shader on l.tex to fill its error texture
	local bestPos = l:calcError(input)
	l:pullMap(input, bestPos)
end

if glush then 
	worldScale = 1.9
	gl.glDepthFunc(gl.GL_LEQUAL)
	function glush.display()
		glreport('display begin')
		gl.glEnable(gl.GL_DEPTH_TEST)

		if doReset then
			l:reset()
			cachedDrawElements = nil
			doReset = false
		end

--[[
		glEnable(GL_TEXTURE_2D)
		glBindTexture(GL_TEXTURE_2D, l.texs:cur().id)
		
		glMatrixMode(GL_MODELVIEW)
		glPushMatrix()
		glTranslate(-5,0,0)
		glScale(2,2,2)

		glColor(1,1,1)
		glBegin(GL_QUADS)
		glTexCoord(0,1)	glVertex(0,0)
		glTexCoord(0,0) glVertex(0,1)
		glTexCoord(1,0) glVertex(1,1)
		glTexCoord(1,1) glVertex(1,0)
		glEnd()
		
		glPopMatrix()
		
		glBindTexture(GL_TEXTURE_2D, 0)
		glDisable(GL_TEXTURE_2D)
--]]

-- [[ draw the source image
		gl.glEnable(gl.GL_TEXTURE_2D)
		src.visibleTex:bind()
		gl.glMatrixMode(gl.GL_MODELVIEW)
		gl.glPushMatrix()
		gl.glScale(worldScale,worldScale,worldScale)
		gl.glColor(1,1,1)

		if showWorldList then
			gl.glCallList(showWorldList)
		else
			showWorldList = gl.glGenList()
			gl.glNewList(showWorldList, gl.GL_COMPILE_AND_EXECUTE)
			local maxi = 20
			local maxj = 40
			for j=0,maxj-1 do
				gl.glBegin(gl.GL_TRIANGLE_STRIP)
				for i=0,maxi do
					local fi = i/maxi
					local fj = j/maxj
					local fj1 = (j+1)/maxj
					local th = fi*2*math.pi
					local ph = (fj-.5)*math.pi
					local ph1 = (fj1-.5)*math.pi
					local costh = math.cos(th)
					local cosph = math.cos(ph)
					local cosph1 = math.cos(ph1)
					gl.glTexCoord(fi,fj1)
					gl.glVertex(cosph1 * math.cos(th), cosph1 * math.sin(th), -math.sin(ph1))
					gl.glTexCoord(fi,fj)
					gl.glVertex(cosph * math.cos(th), cosph * math.sin(th), -math.sin(ph))
				end
				gl.glEnd()
			end
			gl.glEndList()
		end	
		gl.glPopMatrix()
		
		src.visibleTex:unbind()
		gl.glDisable(gl.GL_TEXTURE_2D)
--]]
		
-- [[ now draw the points
		gl.glMatrixMode(gl.GL_MODELVIEW)
		gl.glPushMatrix()
		gl.glScale(2,2,2)

		-- draw the lattice in the points
		do
			-- update it to the texture via pbo's
			gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER, l.vbo)
			gl.glBindTexture(gl.GL_TEXTURE_2D, l.texs:cur().id)
			gl.glGetTexImage(gl.GL_TEXTURE_2D, 0, gl.GL_RGBA, gl.GL_FLOAT, 0)
			gl.glBindTexture(gl.GL_TEXTURE_2D, 0)
			gl.glBindBuffer(gl.GL_PIXEL_PACK_BUFFER, 0)
			
			displayShader:use()
			hsvTex:bind()
			gl.glPointSize(3)
			gl.glColor(0,1,0)

			glreport('draw points begin')
			gl.glEnableClientState(gl.GL_VERTEX_ARRAY)	--setup client state drawing
			gl.glBindBuffer(gl.GL_ARRAY_BUFFER, l.vbo)		--setup the vertex vbo
			gl.glVertexPointer(4, gl.GL_FLOAT, 0, 0)		-- 4th param means cast raw pointer!
			gl.glBindBuffer(gl.GL_ARRAY_BUFFER, 0)		-- cleanup
			gl.glDrawArrays(gl.GL_POINTS, 0, l.width * l.height)
			gl.glDrawArrays(gl.GL_LINE_LOOP, 0, l.width * l.height)
			gl.glDisableClientState(gl.GL_VERTEX_ARRAY)
			glreport('draw points end')
			
			gl.glPointSize(1)
			hsvTex:unbind()
			ShaderProgram.useNone()
		end

		gl.glPopMatrix()
--]]

--[[
		if doDetailUp then
			doDetailUp = false
		end

		if doDetailDown then
			doDetailDown = false
		end
--]]

		if doUpdate or doSingleUpdate then
			if doSingleUpdate then
				doUpdate = false
				doSingleUpdate = false
			end

			gl.glMatrixMode(gl.GL_PROJECTION)
			gl.glLoadIdentity()
			gl.glOrtho(0,1,0,1,-1,1)
			gl.glMatrixMode(gl.GL_MODELVIEW)
			gl.glLoadIdentity()
			
			update()
		end
		glreport('display end')
	end
else
	repeat
		update()
		print(l:tostring())
	until false
end
